@@ -28,7 +28,7 @@
 #include <sys/mman.h>
 #include "hash.h"
 #include "linuxelf.h"
-#include "elf_target.h"
+#include "sysdep.h"
 #ifdef IBCS_COMPATIBLE
 #include <ibcs/unistd.h>
 #else
@@ -40,66 +40,72 @@
 #include "../config.h"
 #endif
 
+extern char *_dl_progname;
+
 #ifdef USE_CACHE
 
-/* Look up a library in the cache /etc/ld.so.cache */
-static caddr_t  _dl_lookup_cache (char *libname) {
-  int i, fd;
+static caddr_t _dl_cache_addr = NULL;
+static size_t _dl_cache_size = 0;
+
+int _dl_map_cache(void)
+{
+  int fd;
   struct stat st;
-  caddr_t cache_addr = NULL;
-  short libnamelen = _dl_strlen (libname);
   header_t *header;
-  libentry_t *libent;
 
-  if (_dl_stat (LDSO_CACHE, &st) || (fd = _dl_open (LDSO_CACHE, O_RDONLY)) < 0)
-    return NULL;
+  if (_dl_cache_addr != NULL)
+    return -1;
 
-  if ((cache_addr = (caddr_t) _dl_mmap (0, st.st_size, PROT_READ, MAP_SHARED, fd, 0)) ==
-      (caddr_t) - 1) {
-    _dl_fdprintf(2, "can't map cache\n");
-    _dl_close (fd);
-    return NULL;
+  if (_dl_stat(LDSO_CACHE, &st) || (fd = _dl_open(LDSO_CACHE, O_RDONLY)) < 0)
+  {
+    _dl_fdprintf(2, "%s: can't open cache '%s'\n", _dl_progname, LDSO_CACHE);
+    _dl_cache_addr = (caddr_t)-1; /* so we won't try again */
+    return -1;
   }
+
+  _dl_cache_size = st.st_size;
+  _dl_cache_addr = (caddr_t)_dl_mmap(0, _dl_cache_size, PROT_READ,
+				     MAP_SHARED, fd, 0);
   _dl_close (fd);
+  if (_dl_cache_addr == (caddr_t)-1)
+  {
+    _dl_fdprintf(2, "%s: can't map cache '%s'\n", _dl_progname, LDSO_CACHE);
+    return -1;
+  }
 
-  if (_dl_memcmp (((header_t *) cache_addr)->magic, LDSO_CACHE_MAGIC,
-		   LDSO_CACHE_MAGIC_LEN)) {
-    _dl_fdprintf(2, "cache is corrupt\n");
+  header = (header_t *)_dl_cache_addr;
+
+  if (_dl_memcmp(header->magic, LDSO_CACHE_MAGIC, LDSO_CACHE_MAGIC_LEN))
+  {
+    _dl_fdprintf(2, "%s: cache '%s' is corrupt\n", _dl_progname, LDSO_CACHE);
     goto fail;
   }
-  if (_dl_memcmp (((header_t *) cache_addr)->version, LDSO_CACHE_VER,
-		   LDSO_CACHE_VER_LEN)) {
-    _dl_fdprintf(2, "cache has wrong version - expecting %s\n", 
-		LDSO_CACHE_VER);
+
+  if (_dl_memcmp (header->version, LDSO_CACHE_VER, LDSO_CACHE_VER_LEN))
+  {
+    _dl_fdprintf(2, "%s: cache '%s' has wrong version\n", _dl_progname, LDSO_CACHE);
     goto fail;
   }
-  for (i = 0,
-       header = (header_t *) cache_addr,
-       libent = (libentry_t *) (cache_addr + sizeof (header_t));
-       cache_addr && i < header->nlibs; i++)
-    if (libent[i].flags == LIB_ELF &&
-	(libnamelen == (libent[i].libnamelen & 0xff) ||
-	libnamelen == (libent[i].libnamelen >> 8)) &&
-	!_dl_memcmp (libname, cache_addr + header->liboffset +
-		      libent[i].liboffset, libnamelen)
-	) {
-      char *ret = _dl_malloc ((libent[i].libnamelen & 0xff) +
-			      libent[i].dirnamelen + 2);
-      _dl_memcpy (ret,
-		   cache_addr + header->diroffset +
-		   libent[i].diroffset, libent[i].dirnamelen);
-      *(ret + libent[i].dirnamelen) = '/';
-      _dl_memcpy (ret + libent[i].dirnamelen + 1,
-		   cache_addr + header->liboffset +
-		   libent[i].liboffset, libent[i].libnamelen & 0xff);
-      _dl_munmap (cache_addr, st.st_size);
-      return ret;
-    }
+
+  return 0;
+
 fail:
-  _dl_munmap (cache_addr, st.st_size);
-  return NULL;
-
+  _dl_munmap(_dl_cache_addr, _dl_cache_size);
+  _dl_cache_addr = (caddr_t)-1;
+  return -1;
 }
+
+int _dl_unmap_cache(void)
+{
+  if (_dl_cache_addr == NULL || _dl_cache_addr == (caddr_t)-1)
+    return -1;
+
+  _dl_munmap (_dl_cache_addr, _dl_cache_size);
+  _dl_cache_addr = NULL;
+
+  return 0;
+}
+
 #endif
 
 /*
@@ -123,29 +129,35 @@ struct elf_resolve * _dl_load_shared_library(struct dyn_elf * dpnt,
     pnt++;
   }
 
-  if(libname == full_libname)
-    for(; dpnt; dpnt = dpnt->next) {
-      tpnt = dpnt->dyn;
-      pnt1 = (char *) tpnt->dynamic_info[DT_RPATH];
-      if(pnt1) {
-	pnt1 += (unsigned int) tpnt->loadaddr + tpnt->dynamic_info[DT_STRTAB];
-	while(*pnt1){
-	  pnt2 = mylibname;
-	  while(*pnt1 && *pnt1 != ':') *pnt2++ = *pnt1++;
-	  if(pnt2[-1] != '/') *pnt2++ = '/';
-	  pnt = libname;
-	  while(*pnt) *pnt2++  = *pnt++;
-	  *pnt2++ = 0;
-	  tpnt1 = _dl_load_elf_shared_library(mylibname, 0);
-	  if(tpnt1) return tpnt1;
-	  if(*pnt1 == ':') pnt1++;
-	}
-      }
-    }
+ /* If the filename has any '/', try it straight and leave it at that.
+     For IBCS2 compatibility under linux, we substitute the string 
+     /usr/i486-sysv4/lib for /usr/lib in library names. */
 
+  if (libname != full_libname) {
+#ifdef IBCS_COMPATIBLE
+    if (_dl_strncmp(full_libname, "/usr/lib/", 9) != 0) {
+      tpnt1 = _dl_load_elf_shared_library(full_libname, 0);
+    } else {
+      pnt1 = "/usr/i486-sysv4/lib/";
+      pnt = mylibname;
+      while(*pnt1) *pnt++ = *pnt1++;
+      pnt1 = full_libname + 9;
+      while(*pnt1) *pnt++ = *pnt1++;
+      *pnt++ = 0;
+      tpnt1 = _dl_load_elf_shared_library(mylibname, 0);
+    }
+#else
+    tpnt1 = _dl_load_elf_shared_library(full_libname, 0);
+#endif
+    if (tpnt1)
+      return tpnt1;
+    goto goof;
+  }
+
+  /* Check in LD_{ELF_}LIBRARY_PATH, if specified and allowed */
   pnt1 = _dl_library_path;
-  if(*pnt1 && libname == full_libname) {
-    while(*pnt1){
+  if (pnt1 && *pnt1) {
+    while (*pnt1) {
       pnt2 = mylibname;
       while(*pnt1 && *pnt1 != ':' && *pnt1 != ';') *pnt2++ = *pnt1++;
       if(pnt2[-1] != '/') *pnt2++ = '/';
@@ -158,55 +170,45 @@ struct elf_resolve * _dl_load_shared_library(struct dyn_elf * dpnt,
     }
   }
 
-  /* Still no luck.  Try one last possibility */
+#ifdef USE_CACHE
+  if (_dl_cache_addr != NULL && _dl_cache_addr != (caddr_t)-1)
+  {
+    int i;
+    header_t *header = (header_t *)_dl_cache_addr;
+    libentry_t *libent = (libentry_t *)&header[1];
+    char *strs = (char *)&libent[header->nlibs];
 
-  /* If the filename has any '/', try it straight and leave it at that.  For
-   IBCS2 compatibility under linux, we substitute the string /usr/i486-sysv4/lib for
-   /usr/lib in library names. */
-
-  if(libname != full_libname) {
-#ifdef IBCS_COMPATIBLE
-    if(_dl_strncmp(full_libname, "/usr/lib/", 9) != 0) {
-	tpnt1 = _dl_load_elf_shared_library(full_libname, 0);
-    } else {
-      pnt1 = "/usr/i486-sysv4/lib/";
-      pnt = mylibname;
-      while(*pnt1) *pnt++ = *pnt1++;
-      pnt1 = full_libname + 9;
-      while(*pnt1) *pnt++ = *pnt1++;
-      *pnt++ = 0;
-      tpnt1 = _dl_load_elf_shared_library(mylibname, 0);
-    }
-#else
-	tpnt1 = _dl_load_elf_shared_library(full_libname, 0);
-#endif
-	if (!tpnt1) goto goof;
+    for (i = 0; i < header->nlibs; i++)
+    {
+      if (libent[i].flags == LIB_ELF &&
+	  _dl_strcmp(libname, strs+libent[i].sooffset) == 0 &&
+	  (tpnt1 = _dl_load_elf_shared_library(strs+libent[i].liboffset, 0)))
 	return tpnt1;
+    }
+  }
+#endif
+
+  /* Check in rpath directories */
+  for(; dpnt; dpnt = dpnt->next) {
+    tpnt = dpnt->dyn;
+    pnt1 = (char *) tpnt->dynamic_info[DT_RPATH];
+    if(pnt1) {
+      pnt1 += (unsigned int) tpnt->loadaddr + tpnt->dynamic_info[DT_STRTAB];
+      while(*pnt1){
+	pnt2 = mylibname;
+	while(*pnt1 && *pnt1 != ':') *pnt2++ = *pnt1++;
+	if(pnt2[-1] != '/') *pnt2++ = '/';
+	pnt = libname;
+	while(*pnt) *pnt2++  = *pnt++;
+	*pnt2++ = 0;
+	tpnt1 = _dl_load_elf_shared_library(mylibname, 0);
+	if(tpnt1) return tpnt1;
+	if(*pnt1 == ':') pnt1++;
+      }
+    }
   }
 
-#ifndef IBCS_COMPATIBLE
-
-#ifdef USE_CACHE
-  /* Check the cache. We don't know if IBCS should use it
-     yet. */
-  if ((pnt1 = _dl_lookup_cache(libname)))
-    if ((tpnt1 =  _dl_load_elf_shared_library(pnt1,0)))
-      return tpnt1;
-#endif
-
-  /* try "/lib/". */
-  pnt1 = "/lib/";
-  pnt = mylibname;
-  while(*pnt1) *pnt++ = *pnt1++;
-  pnt1 = libname;
-  while(*pnt1) *pnt++ = *pnt1++;
-  *pnt++ = 0;
-
-  tpnt1 = _dl_load_elf_shared_library(mylibname, 0);
-  if(tpnt1) return tpnt1;
-#endif
-
-
+  /* Check in /usr/lib */
 #ifdef IBCS_COMPATIBLE
   pnt1 = "/usr/i486-sysv4/lib/";
 #else
@@ -217,10 +219,22 @@ struct elf_resolve * _dl_load_shared_library(struct dyn_elf * dpnt,
   pnt1 = libname;
   while(*pnt1) *pnt++ = *pnt1++;
   *pnt++ = 0;
-
   tpnt1 = _dl_load_elf_shared_library(mylibname, 0);
-  if(tpnt1) return tpnt1;
+  if (tpnt1) return tpnt1;
   
+#ifndef IBCS_COMPATIBLE
+  /* Check in /lib */
+  /* try "/lib/". */
+  pnt1 = "/lib/";
+  pnt = mylibname;
+  while(*pnt1) *pnt++ = *pnt1++;
+  pnt1 = libname;
+  while(*pnt1) *pnt++ = *pnt1++;
+  *pnt++ = 0;
+  tpnt1 = _dl_load_elf_shared_library(mylibname, 0);
+  if (tpnt1) return tpnt1;
+#endif
+
 goof:
   /* Well, we shot our wad on that one.  All we can do now is punt */
   if (_dl_internal_error_number) _dl_error_number = _dl_internal_error_number;
@@ -243,9 +257,6 @@ struct elf_resolve * _dl_load_elf_shared_library(char * libname, int flag){
   struct elf_phdr * ppnt;
   int piclib;
   char * status;
-#ifdef DEBUG
-  int lastmapped = 0;
-#endif
   int flags;
   char header[4096];
   int dynamic_info[24];
@@ -270,15 +281,16 @@ struct elf_resolve * _dl_load_elf_shared_library(char * libname, int flag){
      * NO!  When we open shared libraries we may search several paths.
      * it is inappropriate to generate an error here.
      */
-    _dl_fdprintf(2, "Unable to open %s.\n", libname);
+    _dl_fdprintf(2, "%s: can't open '%s'\n", _dl_progname, libname);
 #endif
     _dl_internal_error_number = DL_ERROR_NOFILE;
     return NULL;
   }
   zerofile = _dl_open(zfile, O_RDONLY);
   if(zerofile < 0) {
-    _dl_fdprintf(2, "Unable to open /dev/zero.  Please create.\n");
+    _dl_fdprintf(2, "%s: can't open '/dev/zero'\n", _dl_progname);
     _dl_internal_error_number = DL_ERROR_NOZERO;
+    _dl_close(infile);
     return NULL;
   }; 
  
@@ -288,9 +300,10 @@ struct elf_resolve * _dl_load_elf_shared_library(char * libname, int flag){
       epnt->e_ident[1] != 'E' ||
       epnt->e_ident[2] != 'L' ||
       epnt->e_ident[3] != 'F') {
-    _dl_fdprintf(2, libname);
-    _dl_fdprintf(2, " is not an ELF file\n");
+    _dl_fdprintf(2, "%s: '%s' is not an ELF file\n", _dl_progname, libname);
     _dl_internal_error_number = DL_ERROR_NOTELF;
+    _dl_close(infile);
+    _dl_close(zerofile);
     return NULL;
   };
   
@@ -301,8 +314,10 @@ struct elf_resolve * _dl_load_elf_shared_library(char * libname, int flag){
 #endif
       )){
     _dl_internal_error_number = (epnt->e_type != ET_DYN ? DL_ERROR_NOTDYN : DL_ERROR_NOTMAGIC);
-    _dl_fdprintf(2, libname);
-    _dl_fdprintf(2, " is not an ELF executable for " ELF_TARGET "\n");
+    _dl_fdprintf(2, "%s: '%s' is not an ELF executable for " ELF_TARGET "\n",
+		 _dl_progname, libname);
+    _dl_close(infile);
+    _dl_close(zerofile);
     return NULL;
   };
 
@@ -313,7 +328,8 @@ struct elf_resolve * _dl_load_elf_shared_library(char * libname, int flag){
 
     if(ppnt->p_type == PT_DYNAMIC) {
       if (dynamic_addr)
-	_dl_fdprintf(2, "Error - more than one dynamic section\n");
+	_dl_fdprintf(2, "%s: '%s' has more than one dynamic section\n",
+		     _dl_progname, libname);
       dynamic_addr = ppnt->p_vaddr;
       dynamic_size = ppnt->p_filesz;
     };
@@ -346,8 +362,10 @@ struct elf_resolve * _dl_load_elf_shared_library(char * libname, int flag){
 			  flags, zerofile,
 			  0);
   if(_dl_mmap_check_error(status)) {
-    _dl_fdprintf(2, "mmap failed.\n");
+    _dl_fdprintf(2, "%s: can't map '/dev/zero'\n", _dl_progname);
     _dl_internal_error_number = DL_ERROR_MMAP_FAILED;
+    _dl_close(infile);
+    _dl_close(zerofile);
      return NULL;
   };
   libaddr=(unsigned int)status;
@@ -379,8 +397,11 @@ struct elf_resolve * _dl_load_elf_shared_library(char * libname, int flag){
 			  ppnt->p_offset & 0x7ffff000);
 	
 	if(_dl_mmap_check_error(status)) {
-	    _dl_fdprintf(2, "mmap failed.\n");
-	      _dl_internal_error_number = DL_ERROR_MMAP_FAILED;
+	    _dl_fdprintf(2, "%s: can't map '%s'\n", _dl_progname, libname);
+	    _dl_internal_error_number = DL_ERROR_MMAP_FAILED;
+	    _dl_munmap((char *)libaddr, maxvma-minvma);
+	    _dl_close(infile);
+	    _dl_close(zerofile);
 	    return NULL;
 	};
 	
@@ -408,9 +429,12 @@ struct elf_resolve * _dl_load_elf_shared_library(char * libname, int flag){
 			  flags, infile, 
 			  ppnt->p_offset & 0x7ffff000);
       if(_dl_mmap_check_error(status)) {
-	_dl_fdprintf(2, "mmap failed.\n");
-	  _dl_internal_error_number = DL_ERROR_MMAP_FAILED;
-	  return NULL;
+	_dl_fdprintf(2, "%s: can't map '%s'\n", _dl_progname, libname);
+	_dl_internal_error_number = DL_ERROR_MMAP_FAILED;
+	_dl_munmap((char *)libaddr, maxvma-minvma);
+	_dl_close(infile);
+	_dl_close(zerofile);
+	return NULL;
       };
       /* if(libaddr == 0 && piclib) {
 	libaddr = (unsigned int) status;
@@ -436,9 +460,7 @@ struct elf_resolve * _dl_load_elf_shared_library(char * libname, int flag){
   
   if(!dynamic_addr) {
     _dl_internal_error_number = DL_ERROR_NODYNAMIC;
-    _dl_fdprintf(2, libname);
-    _dl_fdprintf(2, " is missing a DYNAMIC section.\n");
-    
+    _dl_fdprintf(2, "%s: '%s' is missing a dynamic section\n", _dl_progname, libname);
     return NULL;
   }
 
@@ -447,6 +469,7 @@ struct elf_resolve * _dl_load_elf_shared_library(char * libname, int flag){
   dynamic_size = dynamic_size / sizeof(struct dynamic);
   _dl_memset(dynamic_info, 0, sizeof(dynamic_info));
   for(i=0; i< dynamic_size; i++){
+    if( dpnt->d_tag > DT_JMPREL ) {dpnt++; continue; }
     dynamic_info[dpnt->d_tag] = dpnt->d_un.d_val;
     if(dpnt->d_tag == DT_TEXTREL ||
        SVR4_BUGCOMPAT) dynamic_info[DT_TEXTREL] = 1;
@@ -457,6 +480,7 @@ struct elf_resolve * _dl_load_elf_shared_library(char * libname, int flag){
      writable before we perform relocations.  Do this now. They get set back
      again later. */
 
+  if (dynamic_info[DT_TEXTREL]) {
   ppnt = (struct elf_phdr *) &header[epnt->e_phoff];
   for(i=0;i < epnt->e_phnum; i++, ppnt++){
     if(ppnt->p_type == PT_LOAD && !(ppnt->p_flags & PF_W))
@@ -464,6 +488,8 @@ struct elf_resolve * _dl_load_elf_shared_library(char * libname, int flag){
 		   (ppnt->p_vaddr & 0xfff) + (unsigned int) ppnt->p_filesz,
 		   PROT_READ | PROT_WRITE | PROT_EXEC);
   }
+  }
+
 
   tpnt = _dl_add_elf_hash_table(libname, (char *) libaddr, dynamic_info, dynamic_addr, 
 			      dynamic_size);
@@ -481,8 +507,7 @@ struct elf_resolve * _dl_load_elf_shared_library(char * libname, int flag){
   
   if(lpnt) {
     lpnt = (int *) (dynamic_info[DT_PLTGOT] + ((int) libaddr));
-    lpnt[1] = (int) tpnt;
-    lpnt[2] = (int) _dl_linux_resolve;
+    INIT_GOT(lpnt, tpnt);
   };
   
   return tpnt;
@@ -518,9 +543,15 @@ int _dl_copy_fixups(struct dyn_elf * rpnt)
   
 /* Use BROKEN_LINKER if the SVr4 linker is being used. */
 #ifndef BROKEN_LINKER
+#ifdef ELF_USES_RELOCA
+  goof += _dl_parse_copy_information(rpnt, tpnt->dynamic_info[DT_RELA],
+					   tpnt->dynamic_info[DT_RELASZ], 0);
+
+#else
   goof += _dl_parse_copy_information(rpnt, tpnt->dynamic_info[DT_REL],
 					   tpnt->dynamic_info[DT_RELSZ], 0);
 
+#endif
 #else
   /* OK, now scan the symbol table for this module */
 
@@ -567,11 +598,7 @@ int _dl_copy_fixups(struct dyn_elf * rpnt)
 	      while(count--) *to++ = *from++;
 	      /* Cannot use memcpy - SVr4 assembler complains about dup label*/
 #if 0
-	      _dl_fdprintf(2, "Global symbol in ");
-	      _dl_fdprintf(2, tpnt1->libname);
-	      _dl_fdprintf(2, " ");
-	      _dl_fdprintf(2, pnt);
-	      _dl_fdprintf(2, "\n");
+	      _dl_fdprintf(2, "Global symbol in %s %s\n", tpnt1->libname, pnt);
 #endif
 	    };
 	  };
