@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1994 Eric Youngdale, Peter MacDonald, David Engel,
+ * Copyright (c) 1994, 1995 Eric Youngdale, Peter MacDonald, David Engel,
  * Hongjiu Lu and Mitch D'Souza
  *
  * All rights reserved.
@@ -120,8 +120,11 @@ static int try_lib(char *argv0, char *buffer, char *dir, char *lib)
     int found;
 
     strcpy(buffer, dir);
-    strcat(buffer, "/");
-    strcat(buffer, lib);
+    if (lib != NULL)
+    {
+        strcat(buffer, "/");
+	strcat(buffer, lib);
+    }
 
     if (!(found = !uselib(buffer)))
     {
@@ -155,8 +158,7 @@ shared_loader(int func, ...)
     char *libdir[MAX_DIRS];
     struct libentry **ptr;
     char *cp;
-    int i, fd, ldd = 0;
-    static char defpath[] = DEFAULT_PATH;
+    int i, ldd = 0;
     va_list ap;
     struct libentry **__SHARED_LIBRARIES__;
     struct fixuplist *_SHARABLE_CONFLICTS__;
@@ -203,8 +205,8 @@ shared_loader(int func, ...)
 #ifndef DEBUG_LIB
 	mapinfo = va_arg(ap, unsigned *);
 	save_mapinfo(mapinfo);
-	argv0 = va_arg(ap, char *);
 #endif
+	argv0 = va_arg(ap, char *);
 	__environ = va_arg(ap, char **);
 	__SHARED_LIBRARIES__ = va_arg(ap, struct libentry **);
 	_SHARABLE_CONFLICTS__ = va_arg(ap, struct fixuplist *);
@@ -257,24 +259,8 @@ shared_loader(int func, ...)
 	}
     }
 
-    /* If the cache exists then we don't add these */
-    if (!(cache_addr = cache_init(LDSO_CACHE, &cache_size))) 
-    {
-	/* big brother wants you to use these also */
-	if ((fd = open(LDSO_CONF, O_RDONLY, 0)) >= 0)
-	{
-	    struct stat stat;
-	    fstat(fd, &stat);
-	    cp = alloca(stat.st_size + 1);
-	    read(fd, cp, stat.st_size);
-	    close(fd);
-	    cp[stat.st_size] = '\0';
-	    ndirs = add_dirs(argv0, ndirs, libdir, cp);
-	}
-
-	/* these might also be useful */
-	ndirs = add_dirs(argv0, ndirs, libdir, defpath);
-    }
+    /* attempt to read and initialize the cache */
+    cache_addr = cache_init(LDSO_CACHE, &cache_size);
 
     /* OK, enough formalities, which libs do you want? */
     for (ptr = __SHARED_LIBRARIES__+2; *ptr; ptr++)
@@ -307,36 +293,25 @@ shared_loader(int func, ...)
 	    /* not there, maybe it's in the cache library list */
 	    if (!found && cache_addr)
 	    {
-		short libnamelen = strlen(cp);
 		header_t *header = (header_t *)cache_addr;
-		libentry_t *libent = (libentry_t *)(cache_addr +
-						    sizeof(header_t));
+		libentry_t *libent = (libentry_t *)&header[1];
+		char *strs = (char *)&libent[header->nlibs];
 
 		for (i = 0; !found && i < header->nlibs; i++)
 		{
 		    if (libent[i].flags == LIB_DLL &&
-			(libnamelen == (libent[i].libnamelen & 0xff) ||
-			libnamelen == (libent[i].libnamelen >> 8)) &&
-			!memcmp(cp, cache_addr + header->liboffset +
-				libent[i].liboffset, libnamelen)
-			)
-			found = try_lib(argv0, buffer, 
-					cache_addr + header->diroffset + 
-					libent[i].diroffset,
-					cache_addr + header->liboffset + 
-					libent[i].liboffset);
-		}
-
-		/* nope, last resort is the cache directory list */
-		if (!found)
-		{
-		    char *dir = cache_addr + header->diroffset;
-
-		    for (i = 0; !found && i < header->ndirs;
-			 i++, dir += strlen(dir) + 1)
-			found = try_lib(argv0, buffer, dir, cp);
+			strcmp(cp, strs + libent[i].sooffset) == 0)
+		        found = try_lib(argv0, buffer, 
+					strs + libent[i].liboffset,
+					NULL);
 		}
 	    }
+
+	    /* if not found yet, then try /usr/lib and /lib */
+	    if (!found)
+		found = try_lib(argv0, buffer, "/usr/lib", cp);
+	    if (!found)
+		found = try_lib(argv0, buffer, "/lib", cp);
 	}
 
 	/* OK, let's take a look and see what we've found */
@@ -450,7 +425,10 @@ static caddr_t cache_init(char *file, size_t *size)
     caddr_t cache_addr;
 
     if (stat(LDSO_CACHE, &st) || (fd = open(LDSO_CACHE, O_RDONLY)) < 0)
+    {
+	fdprintf(2, "%s: can't open cache '%s'\n", LDSO_IMAGE, LDSO_CACHE);
 	return NULL;
+    }
     *size = st.st_size;
     if ((cache_addr = mmap(0, *size, PROT_READ, MAP_SHARED, fd, 0)) ==
 	(caddr_t)-1)
